/*
 *
 *
 */

package corncadger.search;

import corncadger.Path;
import corncadger.Map;
import corncadger.Field;

abstract class Criteria {
    // null if equal, else the better path
    public abstract Path betterPath(Path _a, Path _b, Map _m);
    
    /*
     * here are some static Criterias
     * there is only one instance of each type (subclass)
     *
     */
  
    /* shorter path */
    public static final Criteria shorterPathCriteria 
        = new ShorterPathCriteria();

    public static final Criteria maybeNearerToHomeCriteria
        = new MaybeNearerToHomeCriteria();

    public static final Criteria maybeFarerFromHomeCriteria
        = new MaybeFarerFromHomeCriteria();

    /* has cliff*/
    public static final Criteria lessCliffsCriteria 
        = new LessCliffsCriteria();

    /* has cliff*/
    public static final Criteria moreCliffsCriteria 
        = new MoreCliffsCriteria();

    /* nearer to next cliff */
    public static final Criteria nearerCliffCriteria
        = new NearerCliffCriteria();

    /* number of unknown neighbors */
    public static final Criteria moreUnknownNeighborsCriteria
        = new MoreUnknownNeighborsCriteria();

    /* number of unwatched neighbors of end field */
    public static final Criteria moreUnwatchedNeighborsCriteria
        = new MoreUnwatchedNeighborsCriteria();

    /* number of unknown walls */
    public static final Criteria moreUnknownWallsCriteria
        = new MoreUnknownWallsCriteria();

    /* number of unwatched walls of end field */
    public static final Criteria moreUnwatchedWallsCriteria
        = new MoreUnwatchedWallsCriteria();

    /* nearest to unknown walls */
    public static final Criteria nearestToUnknownWallsCriteria
        = new NearestToUnknownWallsCriteria();

    /* has higher level */
    public static final Criteria higherLevelCriteria 
        = new HigherLevelCriteria();

    /* more fields with unknown number of corn */
    public static final Criteria moreUnknownNumberOfCornCriteria
        = new MoreFieldsWithUnknownNumberOfCornCriteria();

    /* lower number of corn of endfield criteria */
    public static final Criteria lowerNumberOfCornCriteria
        = new LowerNumberOfCornCriteria();
}

/* private subclasses implementing abstact class Criteria */

// the higher value wins
abstract class GreaterValueCriteria extends Criteria {
    abstract int pathValue(Path _a, Map _m);
    public Path betterPath(Path _a, Path _b, Map _m) {
        int va = this.pathValue(_a, _m);
        int vb = this.pathValue(_b, _m);
        if (va > vb) return _a;
        if (vb > va) return _b;
        return null;
    }
}
// the lower value wins
abstract class LowerValueCriteria extends GreaterValueCriteria {
    public Path betterPath(Path _a, Path _b, Map _m) {
        int va = this.pathValue(_a, _m);
        int vb = this.pathValue(_b, _m);
        if (va < vb) return _a;
        if (vb < va) return _b;
        return null;
    }
}

// well .. what u see is what u get ...

class ShorterPathCriteria extends LowerValueCriteria {
    public ShorterPathCriteria() {}
    int pathValue(Path _a, Map _m) { return _a.length();  }
}

class MaybeNearerToHomeCriteria extends LowerValueCriteria {
    public MaybeNearerToHomeCriteria() {}
    int pathValue(Path _a, Map _m) { return _a.end().radius(); }
}

class MaybeFarerFromHomeCriteria extends GreaterValueCriteria {
    public MaybeFarerFromHomeCriteria() {}
    int pathValue(Path _a, Map _m) { return _a.end().radius(); }
}

class LessCliffsCriteria extends LowerValueCriteria {
    public LessCliffsCriteria() {}
    int pathValue(Path _a, Map _m) { return _a.numberOfCliffDowns(_m); }
}

class MoreCliffsCriteria extends GreaterValueCriteria {
    public MoreCliffsCriteria() {}
    int pathValue(Path _a, Map _m) { return _a.numberOfCliffDowns(_m); }
}

class NearerCliffCriteria extends LowerValueCriteria {
    public NearerCliffCriteria() {}
    int pathValue(Path _a, Map _m) { return _a.distanceToFirstCliff(_m); }
}

class MoreUnknownNeighborsCriteria extends LowerValueCriteria {
    public MoreUnknownNeighborsCriteria() {}
    int pathValue(Path _a, Map _m) { return _a.numberOfKnownNeighbors(_m); }
}

class MoreUnwatchedNeighborsCriteria extends LowerValueCriteria {
    public MoreUnwatchedNeighborsCriteria() {}
    int pathValue(Path _a, Map _m) {
        return _m.numberOfWatchedNeighbors(_a.end());
    }
}

class MoreUnwatchedWallsCriteria extends LowerValueCriteria {
    public MoreUnwatchedWallsCriteria() {}
    int pathValue(Path _a, Map _m) { return _m.numberOfKnownWalls(_a.end()); }
}

class MoreUnknownWallsCriteria extends LowerValueCriteria {
    public MoreUnknownWallsCriteria() {}
    int pathValue(Path _a, Map _m) { return _a.numberOfKnownWalls(_m); }
}

class NearestToUnknownWallsCriteria extends LowerValueCriteria {
    public NearestToUnknownWallsCriteria() {}
    int pathValue(Path _a, Map _m) { return _a.stepsTillUnknownWalls(_m); }
}

class HigherLevelCriteria extends GreaterValueCriteria {
    public HigherLevelCriteria() {}
    int pathValue(Path _a, Map _m) { return _a.end().height(); }
}

class MoreFieldsWithUnknownNumberOfCornCriteria extends LowerValueCriteria {
    public MoreFieldsWithUnknownNumberOfCornCriteria() {}
    int pathValue(Path _a, Map _m) {
        return _a.numberOfFieldsWithKnownNumberOfCorn();
    }
}

class LowerNumberOfCornCriteria extends LowerValueCriteria {
    public LowerNumberOfCornCriteria() {}
    int pathValue(Path _a, Map _m) { return _a.end().corn(); }
}
