package frisswienix;

import algds.c2002.Hamster;
import java.util.Vector;

/**
 * Hamster "FrissWieNix"
 * @author Matthias Wuttke
 */

public class FrissWieNix extends Hamster {

    protected HamsterMap map = new HamsterMap();
    protected HamsterCell home = map.getHomeCell();
    protected HamsterLocation location = new HamsterLocation();

    protected boolean dontDropAtFirstStep = false;

    /**
     * Der Hamsterlauf.
     * 1. Alle Zellen erforschen
     *    (Dabei clever Korn mitnehmen...)
     * 2. Alle Krner (die sich lohnen) einsammeln
     * (3. Zurck nach Hause)
     */
    
    public void run() {
	HamsterLog.hamster = this;

	location.cell = home;
	location.orientation = HamsterMap.DIRECTION_NO;

	boolean ready = false;

	/* Erforschungsphase */
	while (!ready) {
	    HamsterLog.log("Erforsche Zelle " + 
			   HamsterLog.putCoords(location.cell));

	    /* 1. Aktuelle Zelle erforschen */
	    location.orientation = 
		location.cell.exploreCell(this, location.orientation);

	    /* 2. Einen Nachbarn erforschen */
	    if (location.cell.hasUnexploredNeighbours())
		exploreNeighbour();

	    /* 3. Ein neues Forschungsziel suchen */
	    else
		ready = !goNextUnexploredCell();
	}

	/* Sammelphase */
	ready = false;
	do {
	    HamsterCell heap = findNearestHeap();
	    boolean cheap = isHeapCheap(heap);
	    if (!cheap && heap != null) {
		heap = findBiggestHeap();
		cheap = isHeapCheap(heap);
	    }

	    if (heap != null && cheap) {
		goTo(heap);
		takeAsMuchAsPossible();
		dontDropAtFirstStep = true;
	    }

	    if (!atHome() && (mouthFull() || !cheap || heap == null))
		goHome();
	    else
		ready = atHome() && (heap == null || !cheap);
	} while (!ready);
    }

    /**
     * Geht auf eine unerforschte Nachbarzelle. Ruft fr diese Zelle nicht
     * exploreCell auf.
     * Nicht aufrufen wenn's keinen Nachbarn mehr zu erforschen gibt.
     */

    public void exploreNeighbour() {
	if (!location.cell.hasUnexploredNeighbours())
	    throw new IllegalArgumentException("keine unerforschten Nachbarn");

	boolean foundDirection = false;
	HamsterCell neighbour = null;
	do {
	    int wall = location.cell.getWall(location.orientation);
	    neighbour = map.getNeighbour(location.cell, location.orientation);
	    if (wall == HamsterCell.WALL_NO && !neighbour.isCellExplored())
		foundDirection = true;
	    else {
		// Hamster dreh' dich!
		turn(TURN_RIGHT);
		location.orientation++;
		if (location.orientation == 6)
		    location.orientation = 0;
	    }
	} while (!foundDirection);

	// Schritt vorwrts
	decideCarryBeforeStep();
	if (!forward())
	    throw new RuntimeException("Kollision!");
	location.cell = neighbour;
    }

    /**
     * Entscheidet unmittelbar vor einem Schritt,
     * ob Korn mitgenommen werden soll.
     */

    public void decideCarryBeforeStep() {
	HamsterCell nextCell = map.getNeighbour(location.cell, 
						location.orientation);

	int thisDistance = location.cell.getHomeDistance();
	if (thisDistance == -1)
	    throw new IllegalArgumentException("Feld nicht erforscht");

	int nextDistance = nextCell.getHomeDistance();
	if (nextDistance == -1) // drfte nie vorkommen
	    nextDistance = thisDistance + 1;

	/* im Zweifelsfall mitnehmen... 
	 seltener Fall wo mitnehmen schlecht und thisDistance = nextDistance:
	 die nchsten Felder sind alle belegt so dass das Korn abgelegt werden
	 muss auf einem entfernteren Feld */
	if (thisDistance >= nextDistance)
	    takeAsMuchAsPossible();
	else if (nextDistance > thisDistance) {
	    /* manchmal mag man es nicht, wenn der Hamster Korn wegtut
	       obwohl das nchste Feld weiter weg ist. nmlich dann, wenn
	       er zwei Haufen einsammelt, die gleichweit von zu Hause weg
	       sind */
	    if (!dontDropAtFirstStep)
		leaveAsMuchAsPossible();
	}

	dontDropAtFirstStep = false;
    }

    /**
     * Nimm soviel Korn mit wie mglich.
     */

    public void takeAsMuchAsPossible() {
	int cornOnField = location.cell.getCorn();
	if (cornOnField == -1)
	    throw new IllegalArgumentException("Feld nicht erforscht");

	if (cornOnField == 0)
	    return;

	int cornMouth = getCarriedCorn();
	int takeAway = cornOnField;

	if (takeAway + cornMouth > CORNCARRY_MAX)
	    takeAway = CORNCARRY_MAX - cornMouth;

	location.cell.setCorn(cornOnField - takeAway);
	take(takeAway);

	HamsterLog.log("Nimm " + takeAway + " Korn von Feld " +
		       HamsterLog.putCoords(location.cell) + " mit.");
    }

    /**
     * Lsst soviel Korn wie mglich auf dem aktuellen Feld liegen.
     */

    public void leaveAsMuchAsPossible() {
	int cornOnField = location.cell.getCorn();
	if (cornOnField == -1)
	    throw new IllegalArgumentException("Feld nicht erforscht");

	int cornMouth = getCarriedCorn();
	if (cornMouth == 0)
	    return;

	int leaveHere = cornMouth;
	if (!atHome())
	    if (cornOnField + leaveHere > CORNFIELD_MAX)
		leaveHere = CORNFIELD_MAX - cornOnField;

	location.cell.setCorn(cornOnField + leaveHere);
	dropHere(leaveHere);
	
	HamsterLog.log("Lasse " + leaveHere + " Korn auf Feld " +
		       HamsterLog.putCoords(location.cell) + " (Entfernung " +
		       location.cell.getHomeDistance() + ") zurck.");
    }

    /**
     * Geht zur nchsten unerforschten Zelle.
     * Liefert true zurck wenn erfolgreich und false
     * wenn es keine unerforschte Zelle mehr gibt.
     */

    public boolean goNextUnexploredCell() {
	/* Flag zurcksetzen */
	map.clearFlags();

	/* Aktuelle Zelle */
	HamsterCell current = location.cell;
	current.setFlag(0);

	/* Erste angetroffene unerforschte Zelle */
	HamsterCell unexplored = null;

	/* Wurde eine unerforschte Zelle gefunden? */
	boolean found = false;

	/* Warteschlange */
	Vector queue = new Vector();
	queue.addElement(current);

	/* Solange bis Schlange leer oder unerforschte Zelle gefunden */
	while (queue.size() > 0 && !found) {
	    current = (HamsterCell) queue.elementAt(0);
	    queue.removeElementAt(0);

	    int dist = current.getFlag() + 1;

	    /* alle Nachbarn einreihen */
	    for (int direction = 0; direction < 6; direction++)
		/* geht's in die Richtung? */
		if (current.getWall(direction) == HamsterCell.WALL_NO) {
		    HamsterCell neighbour = map.getNeighbour(current, 
							     direction);
		    /* gibt es eine Nachbarzelle? ist sie schon erreicht? */
		    if (neighbour != null &&
			neighbour.getFlag() == -1) {
			neighbour.setFlag(dist);
			if (neighbour.isCellExplored())
			    queue.addElement(neighbour);
			else {
			    unexplored = neighbour;
			    found = true;
			}
		    }
		}
	}

	/* kein Weg?? */
	if (!found)
	    return false;

	HamsterLog.log("Unerforschte Zelle gefunden: " + 
		       HamsterLog.putCoords(unexplored));
	    
	/* den Weg zur gefundenen Zelle speichern und loslaufen */
	Vector way = findFlagWay(unexplored);
	HamsterLog.log("Weg dahin hat " + way.size() + " Schritte.");
	goWay(way);

	return true;
    }

    /**
     * Findet den Nachbarn der bergebenen Zelle, der erreichbar
     * ist und dessen Flag "dist" ist. Gibt es mehrere solche
     * Zellen wird die genommen, die nher an Zuhause liegt.
     * (Begnstigt Kornmitnahmeverhalten.)
     */

    public HamsterCell findNeighbour(HamsterCell cell, int dist) {
	int homeDist = 1000;
	HamsterCell found = null;
	for (int i = 0; i < 6; i++) {
	    HamsterCell neighbour = map.getNeighbour(cell, i);
	    if (neighbour != null && neighbour.getFlag() == dist
		&& cell.getWall(i) == HamsterCell.WALL_NO) {
		if (homeDist > neighbour.getHomeDistance() ||
		    found == null) {
		    homeDist = neighbour.getHomeDistance();
		    found = neighbour;
		}
	    }
	}

	return found;
    }

    /**
     * Luft einen Weg. Das erste Element des Vektors ist
     * die aktuelle HamsterCell, das letzte Element die Zielzelle,
     * auf der der Hamster nach dieser Methode steht.
     */

    public void goWay(Vector way) {
	for (int i = 1; i < way.size(); i++) {
	    HamsterCell current = location.cell;
	    HamsterCell next = (HamsterCell)way.elementAt(i);

	    int orientationDest = map.getCellOrientation(current, next);
	    if (orientationDest > location.orientation)
		turn(orientationDest - location.orientation, TURN_RIGHT);
	    else if (orientationDest < location.orientation)
		turn(location.orientation - orientationDest, TURN_LEFT);

	    decideCarryBeforeStep();
	    if (!forward())
		throw new RuntimeException("Kollision!");
	    location.cell = next;
	}
    }

    /**
     * Hamster dreh' Dich!
     * Parameter: Drehrichtung, wie oft drehen
     */

    public void turn(int amount, int direction) {
	for (int i = 0; i < amount; i++) {
	    turn(direction);
	    if (direction == TURN_RIGHT)
		location.orientation++;
	    else
		location.orientation--;
	    if (location.orientation < 0)
		location.orientation = 5;
	    if (location.orientation > 5)
		location.orientation = 0;
	}
    }

    /**
     * Wer hat den Hamster erfunden?
     */

    public String getHamsterAuthor() {
	return "Matthias Wuttke";
    }

    /**
     * Welche Hamsterversion?
     */

    public String getHamsterVersion() {
	return "0.1";
    }

    /**
     * Letzte Hamster-Mutation...
     */

    public String getHamsterDate() {
	return "1. Februar 2002";
    }

    /**
     * Bestimmt, ob der Hamster die Schnauze voll hat.
     */

    public boolean mouthFull() {
	return getCarriedCorn() == CORNCARRY_MAX;
    }

    /**
     * Bestimmt, ob der bergebene Haufen ein Haufen ist, zu dem
     * es sich lohnt, zu gehen.
     * Bercksichtigt dabei
     * - das Fassungsvermgen des Hamstermauls
     * - die aktuelle Position bzw. die Entfernung vom Haufen
     * - die Menge der aufnehmbaren Krner
     * - die Entfernung des Haufens von Zuhause
     */
    
    public boolean isHeapCheap(HamsterCell heap) {
	if (atHome())
	    leaveAsMuchAsPossible();
	else
	    takeAsMuchAsPossible();

	if (heap == null)
	    return false;

	int canTake = heap.getCorn();
	if (canTake > CORNCARRY_MAX - getCarriedCorn())
	    canTake = CORNCARRY_MAX - getCarriedCorn();

	int heapGain = canTake * POINTS_CORN;
	int heapCost = heap.getHomeDistance();

	if (!atHome())
	    heapCost += getDistance(location.cell, heap);
	else
	    heapCost *= 2; 
	// man muss auch zurck

	// nicht bercksichtigt: eine "Rundtour" bei der man mehrere
	// Haufen besucht

	HamsterLog.log("Haufen " + HamsterLog.putCoords(heap) + " kostet " +
		       heapCost + " und bringt ein: " + heapGain);
	
	return heapGain > heapCost;
    }

    /**
     * Geht nach Hause.
     */

    public void goHome() {
	if (!atHome()) {
	    goTo(home);
	    leaveAsMuchAsPossible();
	}
    }

    /**
     * Geht zu einer bestimmten Zelle.
     */

    public void goTo(HamsterCell cell) {
	Vector way = findWay(location.cell, cell);
	goWay(way);
    }

    /**
     * Findet den Weg zwischen zwei Zellen.
     * Rckgabe: ein Vector mit HamsterCells,
     * der erste Eintrag ist cell1, der zweite Eintrag ist cell2.
     * Null wenn kein Weg existiert.
     */

    public Vector findWay(HamsterCell cell1, HamsterCell cell2) {
	/* Flag zurcksetzen */
	map.clearFlags();

	/* Aktuelle Zelle: Start bei cell1 */
	HamsterCell current = cell1;
	current.setFlag(0);

	/* Wurde die Zelle cell2 gefunden? */
	boolean found = false;

	/* Warteschlange */
	Vector queue = new Vector();
	queue.addElement(current);

	/* Solange bis Schlange leer oder Zelle gefunden */
	while (queue.size() > 0 && !found) {
	    current = (HamsterCell) queue.elementAt(0);
	    queue.removeElementAt(0);

	    int dist = current.getFlag() + 1;

	    /* alle Nachbarn einreihen */
	    for (int direction = 0; direction < 6; direction++)
		/* geht's in die Richtung? */
		if (current.getWall(direction) == HamsterCell.WALL_NO) {
		    HamsterCell neighbour = map.getNeighbour(current, 
							     direction);
		    /* gibt es eine Nachbarzelle? ist sie schon erreicht? */
		    if (neighbour != null &&
			neighbour.getFlag() == -1) {
			neighbour.setFlag(dist);
			/* ist die Nachbarzelle die gesuchte? */
			if (neighbour != cell2)
			    queue.addElement(neighbour);
			else
			    found = true;
		    }
		}
	}

	/* kein Weg?? */
	if (!found) {
	    HamsterLog.log("Kein Weg von " + HamsterLog.putCoords(cell1) +
			   " nach " + HamsterLog.putCoords(cell2));
	    return null;
	}

	return findFlagWay(cell2);
    }

    /**
     * Findet, ausgefllte Flags vorausgesetzt, den Weg von
     * der Zelle "destination" zur Zelle mit dem Flag 0.
     */

    public Vector findFlagWay(HamsterCell destination) {
	Vector way = new Vector();
	boolean found = false;
	HamsterCell current = destination;
	while (!found) {
	    way.insertElementAt(current, 0);
	    int dist = current.getFlag() - 1;
	    if (dist == 0) {
		way.insertElementAt(location.cell, 0);
		found = true;
	    } else
		current = findNeighbour(current, dist);

	    if (current == null)
		throw new RuntimeException("kein Nachbar!");
	}
	return way;
    }

    /**
     * Findet die Entfernung zwischen zwei Zellen.
     */

    public int getDistance(HamsterCell cell1, HamsterCell cell2) {
	if (cell1 == cell2)
	    return 0;

	Vector way = findWay(cell1, cell2);
	return way.size() - 1;
    }

    /**
     * Findet den an der aktuellen Position nchstliegenden Haufen.
     * Null: wir haben keine Haufen mehr
     */

    public HamsterCell findNearestHeap() {
	/* Flag zurcksetzen */
	map.clearFlags();

	/* Aktuelle Zelle: Start bei Hamster-Posititon */
	HamsterCell current = location.cell;
	current.setFlag(0);

	/* Warteschlange */
	Vector queue = new Vector();
	queue.addElement(current);

	/* Solange bis Schlange leer */
	while (queue.size() > 0) {
	    current = (HamsterCell) queue.elementAt(0);
	    queue.removeElementAt(0);

	    int dist = current.getFlag() + 1;

	    /* alle Nachbarn einreihen */
	    for (int direction = 0; direction < 6; direction++)
		/* geht's in die Richtung? */
		if (current.getWall(direction) == HamsterCell.WALL_NO) {
		    HamsterCell neighbour = map.getNeighbour(current, 
							     direction);
		    /* gibt es eine Nachbarzelle? ist sie schon erreicht? */
		    if (neighbour != null && neighbour.getFlag() == -1) {
			/* Haufen gefunden? */
			if (neighbour.getCorn() > 0 && neighbour != home)
			    return neighbour;
			else {
			    neighbour.setFlag(dist);
			    queue.addElement(neighbour);
			}
		    }
		}
	}
	
	HamsterLog.log("Alle Haufen eingesammelt.");
	return null; /* kein Haufen mehr */
    }

    /**
     * Findet den grten Haufen.
     * Null: es gibt keinen Haufen mehr.
     */

    public HamsterCell findBiggestHeap() {
	HamsterCell max = null;
	int maxC = MAZELENGTH_MAX * 2 - 1;
	for (int x = 0; x < maxC; x++)
	    for (int y = 0; y < maxC; y++) {
		HamsterCell cell = map.getCell(x, y);
		if (cell.getCorn() > 0 && cell != home) {
		    if (max == null)
			max = cell;
		    else if (max.getCorn() < cell.getCorn())
			max = cell;
		    else if (max.getCorn() == cell.getCorn()) {
			if (max.getHomeDistance() > cell.getHomeDistance())
			    max = cell;
		    }
		}
	    }
		
	return max;
    }

}
