package frisswienix;
import algds.c2002.Hamster;

public class HamsterCell {

    /**
     * Ist hier eine Wand?
     * Ja/Nein/Vielleicht
     */

    public static final int WALL_YES = 1;
    public static final int WALL_NO = 2;
    public static final int WALL_DONTKNOW = 3;

    /**
     * Klassenvariablen.
     */

    protected HamsterMap map = null;
    protected int corn = -1;
    protected int walls[] = new int[6];
    protected int x;
    protected int y;
    protected int homeDistance = -1;
    protected int flag = -1;
    protected boolean hasBeenSeen = false;

    /**
     * Initialisiert diese Zelle als Teil der Karte "map" an
     * der Position x/y.
     */

    public HamsterCell(HamsterMap map, int x, int y) {
	this.map = map;
	for (int i = 0; i < 6; i++)
	    walls[i] = WALL_DONTKNOW;
	this.x = x;
	this.y = y;
    }

    /**
     * Berechnet die Entfernung dieser Zelle zum Zuhause.
     * Der Wert wird gespeichert, muss aber regelmssig
     * (bei jeder neuen Erforschung) aktualisiert werden.
     * Der Hamster knnte einen krzeren Weg entdeckt haben.
     * Von dieser Zelle ist kein Weg nach Hause bekannt: -1
     */

    public int getHomeDistance() {
	return homeDistance;
    }

    /**
     * Setzt die Entfernung vom Zuhause. Wird von
     * HamsterMap.calculateHomeDistances() aufgerufen.
     */

    protected void setHomeDistance(int hd) {
	this.homeDistance = hd;
    }

    /**
     * Liefert zurck, wie viel Korn auf diesem Feld liegt.
     * 0: hier liegt kein Korn
     * -1: nicht erforscht
     */

    public int getCorn() {
	return corn;
    }

    /**
     * Aktualisiert die Kornzahl auf diesem Feld.
     * Wenn man was nimmt/ablegt.
     */

    public void setCorn(int corn) {
	this.corn = corn;
    }

    /**
     * Liefert den "Wand-Status" in der gegebenen Richtung
     * zurck. Siehe Konstanten fr Richtung und Wand.
     */

    public int getWall(int dir) {
	return walls[dir];
    }

    /**
     * Setzt den "Wand-Status" in der gegebenen Richtung.
     */

    public void setWall(int dir, int wall) {
	walls[dir] = wall;
    }

    /**
     * Setzt den Has Been Seen-Status.
     */

    public void setHasBeenSeen(boolean hasBeenSeen) {
	this.hasBeenSeen = hasBeenSeen;
    }

    /**
     * Ist diese Zelle schon gesehen worden?
     */

    public boolean getHasBeenSeen() {
	return hasBeenSeen;
    }

    /**
     * Erforscht dieses Feld und umliegende Wnde, soweit sie
     * es noch nicht sind. Nur wenn der Hamster auf diesem
     * Feld steht! Wenn sich etwas neues ergibt (neue Wnde)
     * wird die Home Distance aller Zellen neu berechnet.
     * @param hamster der Hamser der sich dreht
     * @param orientation wohin guckt der Hamster gerade
     * @return die neue Blickrichtung des Hamsters
     */

    public int exploreCell(Hamster hamster, int orientation) {
	// das aktuelle Feld erforschen
	setHasBeenSeen(true);
	setCorn(hamster.getAvailableCorn());
	if (corn > 0)
	    HamsterLog.log("Auf dem Feld " + HamsterLog.putCoords(this) +
			   " liegt Korn: " + corn);

	// gibt's Wnde, die man sich angucken sollte?
	if (!hasUnexploredWalls())
	    return orientation;

	for (int i = 0; i < 6; i++) {
	    if (getWall(orientation) == WALL_DONTKNOW) {
		// diese Wand erforschen
		int look = hamster.lookForward();
		if ((look & hamster.WALLMASK) == hamster.WALL) {
		    /* Hamster steht vor Wand */
		    walls[orientation] = WALL_YES;
		    setNeighbourWall(orientation, WALL_YES);
		} else {
		    /* Vorm Hamster ist keine Wand */
		    walls[orientation] = WALL_NO;
		    setNeighbourWall(orientation, WALL_NO);

		    /* Ist auf dem angrenzenden Feld Korn? */
		    if ((look & hamster.CORNMASK) != hamster.CORN) {
			HamsterCell neighbour = map.getNeighbour(this, 
			       orientation);

			if (neighbour == null)
			    throw new IllegalArgumentException("Karte offen!");

			neighbour.setCorn(0);
			neighbour.setHasBeenSeen(true);
		    }
		}

		// keine weiteren unerforschten Wnde? --> fertig
		if (!hasUnexploredWalls()) {
		    map.calculateHomeDistances();
		    return orientation;
		}
	    }

	    // Hamster dreh' dich!
	    hamster.turn(hamster.TURN_RIGHT);
	    orientation++;
	    if (orientation == 6)
		orientation = 0;
	}

	map.calculateHomeDistances();
	return orientation;
    }

    /**
     * Synchronisiert diese Wand mit der Wand der Nachbarzelle.
     * Geteilte Wnde sozusagen...
     */

    public void setNeighbourWall(int orientation, int wall) {
	HamsterCell neighbour = map.getNeighbour(this, orientation);
	if (neighbour == null)
	    return;

	int opposite = map.getOppositeDirection(orientation);
	neighbour.setWall(opposite, wall);
    }

    /**
     * Ist diese Zelle fertig erforscht?
     * D.h. weiss der Hamster wieviel Korn hier ist?
     * Weiss er ber alle Wnde bescheid?
     * Wnde interessieren nicht, wenn hier sowieso kein Korn
     * ist und auch keine Nachbarzellen mehr zu entdecken sind.
     */

    public boolean isCellExplored() {
	boolean wallsExplored = !hasUnexploredWalls();
	boolean cornExplored = getCorn() != -1;

	if (!cornExplored)
	    return false;

	if (wallsExplored && cornExplored)
	    return true;

	/* Wnde noch nicht erforscht. */
	for (int i = 0; i < 6; i++) {
	    HamsterCell neighbour = map.getNeighbour(this, i);
	    if (neighbour != null && getWall(i) != WALL_YES)
		if (neighbour.getHasBeenSeen() == false ||
		    neighbour.getCorn() != 0)
		    return false;
	}

	return hasBeenSeen;

	/*
	boolean wallsExplored = !hasUnexploredWalls();
	boolean cornExplored = getCorn() != -1;
	   
	return wallsExplored && cornExplored;
	*/
    }

    /**
     * Hat diese Zelle unerforschte Wnde?
     */

    public boolean hasUnexploredWalls() {
	for (int i = 0; i < 6; i++)
	    if (getWall(i) == WALL_DONTKNOW)
		return true;
	return false;
    }

    /**
     * Hat diese Zelle unerforschte Nachbarn?
     * Liefert auch true, wenn eine Wand noch nicht
     * erforscht ist.
     */

    public boolean hasUnexploredNeighbours() {
	for (int i = 0; i < 6; i++) {
	    int wall = getWall(i);
	    if (wall == WALL_NO) {
		HamsterCell neighbour = map.getNeighbour(this, i);
		if (neighbour == null)
		    throw new IllegalArgumentException("Offene Karte!");
		if (neighbour != null && !neighbour.isCellExplored())
		    return true;
	    } else if (wall == WALL_DONTKNOW)
		return true;
	}
	return false;
    }

    /**
     * X-Koordinate dieser Zelle.
     */

    public int getX() {
	return x;
    }

    /**
     * Y-Koordinate dieser Zelle.
     */

    public int getY() {
	return y;
    }

    /**
     * Ein Nachbarn dieser Zelle.
     */

    public HamsterCell getNeighbour(int direction) {
	return map.getNeighbour(this, direction);
    }

    /**
     * Die Karte auf der diese Zelle liegt.
     */

    public HamsterMap getMap() {
	return map;
    }

    /**
     * Holt das Flag. Das Flag kann von Algorithmen, die
     * mit Zellen arbeiten, frei benutzt werden.
     */

    public int getFlag() {
	return flag;
    }

    /**
     * Setzt das Flag. Das Flag kann von Algorithmen, die
     * mit Zellen arbeiten, frei benutzt werden.
     */

    public void setFlag(int flag) {
	this.flag = flag;
    }

    /**
     * Lscht das Flag. Das Flag kann von Algorithmen, die
     * mit Zellen arbeiten, frei benutzt werden.
     */

    public void clearFlag() {
	setFlag(-1);
    }

}
