package h2d2c;
import h2d2c.util.*;

//---------------------------------------------------------------------
//----- Klasse Cell
//---------------------------------------------------------------------
/**
 * Ein einzelnes Feld in der Karte.
 * @author Christian Semrau
 */
class Cell implements H2D2Constants {

  /** Koordinate des Feldes. */
  final IntPoint position;

  /** Getreide auf dem Feld. 0 = leer; -1, -2 = unbekannt. */
  private int corn;

  /** Waende und Klippen. */
  private final int[] border;

  /** Status des Feldes: Ungesehen, gesehen, betreten. */
  private int state;

  /** Der sichere Heimweg nur ueber bekannte Kanten. */
  final Way wayhome;
  /** Der theoretische Heimweg, wenn alle unbekannten Kanten frei waeren. */
  final Way waymin;

  /** Fuer Breitensuchen. */
  final Way knownway, possibleway;

  /** Nachbarfelder. */
  private Cell[] linx;

//---------------------------------------------------------------------
//----- Konstruktor
//---------------------------------------------------------------------

  Cell(IntPoint position) {
    this.position = position;
    this.corn = UNKNOWNCORNSTATE;
    this.border = new int[NUMDIR];
    this.state = NOTSEEN;
    this.linx = new Cell[NUMDIR];

    for(int dir=0; dir<NUMDIR; dir++) setUnknown(dir);

    wayhome = new Way();
    waymin = new Way();

    knownway = new Way();
    possibleway = new Way();
  }

//---------------------------------------------------------------------
//----- Methoden zum Umgang mit der Position
//---------------------------------------------------------------------

  IntPoint getPosition() {
    return position;
  }

//---------------------------------------------------------------------
//----- Methoden zum Umgang mit Getreide
//---------------------------------------------------------------------

  /** Setzt das Getreide.
   * @param newCorn
   */
  void setCorn(int newCorn) {
    corn = newCorn;
  }

  /** Fuegt Getreide hinzu oder entfernt es.
   * @param diffCorn groesser als 0: hinzufuegen, kleiner als 0: entfernen.
   */
  void changeCorn(int diffCorn) {
    if (corn>=0) {
      corn = Math.max(0, corn+diffCorn);
    } else
    if (diffCorn<0)
      corn = UNKNOWNCORNSTATE;
    else
    if (diffCorn>0)
      corn = UNKNOWNCORNAMOUNT;
  }

  /** Gibt den Getreide-Vorrat des Feldes.
   */
  int getCorn() {
    return corn;
  }

//---------------------------------------------------------------------
//----- Methoden zum Umgang mit Waenden und Klippen
//---------------------------------------------------------------------

  /** Setzt eine unbekannte Seitenart.
   * @param dir
   */
  void setUnknown(int dir) {
    border[dir] = CUNKNOWN;
  }
  /** Gibt an, ob die Art der Seite bekannt ist.
   * @param dir
   */
  boolean isKnown(int dir) {
    return border[dir] != CUNKNOWN;
  }

  /** Gibt an, ob die Art irgendeiner Seite unbekannt ist.
   */
  boolean isAnyUnknown() {
    for (int dir = MINDIR; dir<NUMDIR; dir++)
      if (!isKnown(dir)) return true;
    return false;
  }

  /** Setzt einen ebenen Weg.
   * @param dir
   */
  void setFree(int dir) {
    border[dir] = CFREE;
  }
  /** Gibt an, ob eine ebener Weg vorhanden ist.
   * @param dir
   */
  boolean isFree(int dir) {
    return border[dir] == CFREE;
  }

  /** Setzt eine Wand.
   * @param dir
   */
  void setWall(int dir) {
    border[dir] = CWALL;
  }
  /** Gibt an, ob eine Wand vorhanden ist.
   * @param dir
   */
  boolean isWall(int dir) {
    return border[dir] == CWALL;
  }

  /** Setzt eine aufsteigende Klippe.
   * @param dir
   */
  void setCliffUp(int dir) {
    border[dir] = CUP;
  }
  
  /** Gibt an, ob eine aufsteigende Klippe vorhanden ist.
   * @param dir
   */
  boolean isCliffUp(int dir) {
    return border[dir] == CUP;
  }

  /** Setzt eine absteigende Klippe.
   * @param dir
   */
  void setCliffDown(int dir) {
    border[dir] = CDOWN;
  }
  
  /** Gibt an, ob eine absteigende Klippe vorhanden ist.
   * @param dir
   */
  boolean isCliffDown(int dir) {
    return border[dir] == CDOWN;
  }

//---------------------------------------------------------------------
//----- Methoden zum Umgang mit dem Status
//---------------------------------------------------------------------

  /** Setzt den Status.
   * @param newState
   */
  void setState(int newState) {
    if (newState==NOTSEEN || newState==JUSTSEEN || newState==ENTERED){
      state = newState;
    }else
      throw new IllegalArgumentException("setState("+newState+")");
  }

  /** Erhoeht den Status auf den angegebenen Wert.
   * @param newState
   */
  void incState(int newState) {
    if (newState==NOTSEEN || newState==JUSTSEEN || newState==ENTERED){
      if (state < newState)
        state = newState;
    }else
      throw new IllegalArgumentException("incState("+newState+")");
  }

  /** Gibt den Status des Feldes.
   */
  int getState() {
    return state;
  }

  /** Gibt an, ob irgendein Nachbar noch nicht gesehen wurde.
   */
  boolean hasNotseenNeighbor() {
    for (int dir = MINDIR; dir<NUMDIR; dir++)
      if (linx[dir].getState()==NOTSEEN) return true;
    return false;
  }

  /** Gibt an, wieviele Nachbarn noch nicht gesehen wurden.
   */
  int countNotseenNeighbor() {
    int count = 0;
    for (int dir = MINDIR; dir<NUMDIR; dir++)
      if (linx[dir].getState()==NOTSEEN) count++;
    return count;
  }

  /** Gibt an, wieviele Nachbarn nur gesehen wurden.
   */
  int countJustseenNeighbor() {
    int count = 0;
    for (int dir = MINDIR; dir<NUMDIR; dir++)
      if (linx[dir].getState()==JUSTSEEN) count++;
    return count;
  }

  /** Gibt an, wieviele Seiten noch unbekannt sind.
   */
  int countUnknownLinks() {
    int count = 0;
    for (int dir = MINDIR; dir<NUMDIR; dir++)
      if (!isKnown(dir)) count++;
    return count;
  }

//---------------------------------------------------------------------
//----- Methoden zum Umgang mit Nachbarn
//---------------------------------------------------------------------

  /** Setzt einen Nachbarn.
   * @param direction Richtung in der die Nachbarzelle liegt
   * @param neighbor die Nachbarzelle
   */
  void setLink(int direction, Cell neighbor) {
    linx[direction] = neighbor;
  }

  /** Liefert einen Nachbarn.
   * @param direction Richtung in der die Nachbarzelle liegt
   * @param neighbor die Nachbarzelle
   */
  Cell getLink(int direction) {
    return linx[direction];
  }

  public String toString() {
    return "Cell["+position+"]";
  }

  public String fullDebug() {
    String l = "", b = "";
    for(int i=0; i<NUMDIR; i++)
      if (linx[i]==null) l += "-"; else l += "L";
    for(int i=0; i<NUMDIR; i++)
      b += BORDERSTRING[border[i]]+",";
    return "Cell["+position+";b:"+b+";c:"+corn+";"+l+"]";
  }

} // Cell

