package h2d2c;
import java.util.LinkedList;
import h2d2c.util.*;

//---------------------------------------------------------------------
//----- Klasse Map
//---------------------------------------------------------------------
/**
 * @author Christian Semrau
 */
class Map implements H2D2Constants {
  /** Die Felder aus denen das Labyrinth besteht. */
  Cell[][] cells;

  /** Seitenlaenge. */
  int size;

  /** Zeilenzahl = Seitenlaenge * 2 - 1. */
  int len;

  /** Koordinate des Heimatfeldes. */
  IntPoint homePoint;

  /** Zelle des Heimatfeldes. */
  Cell homecell;

//---------------------------------------------------------------------
//----- Konstruktor
//---------------------------------------------------------------------

  Map() {
    initialize();
  }

  private void initialize() {
    size = MAZELENGTH_MAX+1;  // ein Feld Toleranz
    len = size*2-1;
    cells = new Cell[len][len];
    initCells();
    homePoint = new IntPoint(size-1, size-1);
    homecell = getCell(homePoint);
    clearWayMin();
  }

  public void initCells() {
    // Felder erzeugen
    for(int i=0; i<len; i++)
    for(int j=0; j<len; j++)
      cells[i][j] = new Cell(new IntPoint(i,j));

    // Felder verlinken
    for(int i=0; i<len; i++)
    for(int j=0; j<len; j++)
    for(int dir=0; dir<NUMDIR; dir++)
      cells[i][j].setLink(dir, getNeighbor(i,j,dir));
    for(int i=0; i<len; i++) {
      cells[i][0].setWall(N);
      cells[i][0].setWall(NE);
      cells[i][len-1].setWall(S);
      cells[i][len-1].setWall(SW);
      cells[0][i].setWall(NW);
      cells[0][i].setWall(SW);
      cells[len-1][i].setWall(SE);
      cells[len-1][i].setWall(NE);
    }
  }

  public void clearKnownWay() {
    Cell c;
    for(int i=0; i<len; i++)
    for(int j=0; j<len; j++)
    if ((c=getCell(i,j))!=null){
      c.knownway.clear();
    }
  }

  public void clearPossibleWay() {
    Cell c;
    for(int i=0; i<len; i++)
    for(int j=0; j<len; j++)
    if ((c=getCell(i,j))!=null){
      c.possibleway.clear();
    }
  }

  public void clearWayMin() {
    Cell cell;
    int c = homecell.position.x;
    int d = homecell.position.y;
    for(int a=0; a<len; a++)
    for(int b=0; b<len; b++)
    if ((cell=getCell(a,b))!=null){
      int dist, dir;
      int dx = Math.abs(a-c), dy = Math.abs(b-d);
      dist = ((a<=c && b>=d) || (a>=c && b<=d)
        ? Math.max(dx, dy)
        : dx + dy);
      dir = (a==c
        ?(b>d ? N : b<d ? S : -1)
        :a < c
        ?(b > d ? NE : SE)
        :(b < d ? SW : NW)
      );
      
      cell.waymin.set(dist, dir);
    }
  }

//---------------------------------------------------------------------
//----- Methoden zum Umgang mit Zellen
//---------------------------------------------------------------------

  private void setCell(IntPoint p, Cell c) {
    setCell(p.x, p.y, c);
  }

  private void setCell(int x, int y, Cell c) {
    cells[x][y] = c;
  }

  Cell getCell(IntPoint p) {
    int x=p.x, y=p.y;
    if (!posOK(x,y)) return null;
    if (cells[x][y]==null)
      cells[x][y] = new Cell(p);
    return cells[x][y];
  }

  Cell getCell(int x, int y) {
    if (!posOK(x,y)) return null;
    if (cells[x][y]==null)
      cells[x][y] = new Cell(new IntPoint(x,y));
    return cells[x][y];
  }

  Cell getNeighbor(IntPoint p, int dir) {
    return getNeighbor(p.x, p.y, dir);
  }

  Cell getNeighbor(int x, int y, int dir) {
    return getCell(x+WALKLUT[dir][X], y+WALKLUT[dir][Y]);
  }

  boolean posOK(int x, int y) {
    return (x>=0 && y>=0 && x<len && y<len);
  }

//---------------------------------------------------------------------
//----- Methoden zum Umgang mit Waenden und Klippen
//---------------------------------------------------------------------

  /** Entfernt eine Wand. */
  void setFree(IntPoint p, int dir) {
    setFree(p.x, p.y, dir);
  }
  /** Setzt eine Wand. */
  void setWall(IntPoint p, int dir) {
    setWall(p.x, p.y, dir);
  }
  /** Setzt eine aufsteigende Klippe. */
  void setCliffUp(IntPoint p, int dir) {
    setCliffUp(p.x, p.y, dir);
  }
  /** Setzt eine absteigende Klippe. */
  void setCliffDown(IntPoint p, int dir) {
    setCliffDown(p.x, p.y, dir);
  }
  /** Gibt an, ob die Art der Seite bekannt ist. */
  boolean isKnown(IntPoint p, int dir) {
    return isKnown(p.x, p.y, dir);
  }
  /** Gibt an, ob ein ebener Weg vorhanden ist. */
  boolean isFree(IntPoint p, int dir) {
    return isFree(p.x, p.y, dir);
  }
  /** Gibt an, ob eine Wand vorhanden ist. */
  boolean isWall(IntPoint p, int dir) {
    return isWall(p.x, p.y, dir);
  }
  /** Gibt an, ob eine aufsteigende Klippe vorhanden ist. */
  boolean isCliffUp(IntPoint p, int dir) {
    return isCliffUp(p.x, p.y, dir);
  }
  /** Gibt an, ob eine absteigende Klippe vorhanden ist. */
  boolean isCliffDown(IntPoint p, int dir) {
    return isCliffDown(p.x, p.y, dir);
  }

  void setFree(int x, int y, int dir) {
    Cell c = getCell(x,y);
    Cell n = getNeighbor(x,y,dir);
    if (c!=null && n!=null) {
      c.setFree(dir); n.setFree(ANTIDIR[dir]);
    }else{
      setWall(x,y,dir);
    }
  }

  void setWall(int x, int y, int dir) {
    Cell c = getCell(x,y);
    Cell n = getNeighbor(x,y,dir);
    if (c!=null) { c.setWall(dir); }
    if (n!=null) { n.setWall(ANTIDIR[dir]); }
  }

  void setCliffUp(int x, int y, int dir) {
    Cell c = getCell(x,y);
    Cell n = getNeighbor(x,y,dir);
    if (c!=null && n!=null) {
      c.setCliffUp(dir); n.setCliffDown(ANTIDIR[dir]);
    }else
      setWall(x,y,dir);
  }

  void setCliffDown(int x, int y, int dir) {
    Cell c = getCell(x,y);
    Cell n = getNeighbor(x,y,dir);
    if (c!=null && n!=null) {
      c.setCliffDown(dir); n.setCliffUp(ANTIDIR[dir]);
    }else
      setWall(x,y,dir);
  }

  boolean isKnown(int x, int y, int dir) {
    Cell c = getCell(x,y);
    return (c!=null) && (c.isKnown(dir));
  }

  boolean isFree(int x, int y, int dir) {
    Cell c = getCell(x,y);
    return (c!=null) && (c.isFree(dir));
  }

  boolean isWall(int x, int y, int dir) {
    Cell c = getCell(x,y);
    return (c!=null) && (c.isWall(dir));
  }
  
  boolean isCliffUp(int x, int y, int dir) {
    Cell c = getCell(x,y);
    return (c!=null) && (c.isCliffUp(dir));
  }

  boolean isCliffDown(int x, int y, int dir) {
    Cell c = getCell(x,y);
    return (c!=null) && (c.isCliffDown(dir));
  }

  public static boolean DEBUGWAYHOME = false;
  public static boolean DEBUGWAYMIN = true;

  void updateWayHome(Cell start) {
    if (DEBUGWAYHOME && H2D2.currenthamster.verbose())
    H2D2.currenthamster.logln("map.updateWayHome("+start+") entered.");

    PriorityQueue queue = new PriorityQueue(CompareWayHome.COMPARATOR);

    queue.addFirst(start);
    while (!queue.isEmpty()) {
      Cell c = (Cell)queue.removeMin();

      if (DEBUGWAYHOME && H2D2.currenthamster.verbose())
      H2D2.currenthamster.logln("Work at "+c+", home: "+c.wayhome);

      // eigene Heimatentfernung
      int mydist = c.wayhome.distance;
      int mydir = c.wayhome.direction;
      int newdist = INFINITEDISTANCE;
      int newdir = -1;

      // ermittelt die kleinste Heimatentfernung eines Nachbarfeldes
      for(int dir = MINDIR; dir < NUMDIR; dir++) {
        if (c.isWall(dir)) continue;
        if (c.isCliffUp(dir)) continue;
        if (!c.isKnown(dir)) continue;
        Cell n = c.getLink(dir);
        if (n==null) continue;
        int ndist = n.wayhome.distance;
        //int ndir = n.wayhome.direction;
        if (ndist < newdist) {
          newdist = ndist; newdir = dir;
        }
      }

      // neue Heimatentfernung kleiner als aktuelle
      if (newdist+1 < mydist) {

        if (DEBUGWAYHOME && H2D2.currenthamster.verbose())
        H2D2.currenthamster.log("Update from "+c.wayhome);

        mydist = c.wayhome.distance = newdist+1;
        mydir = c.wayhome.direction = newdir;

        if (DEBUGWAYHOME && H2D2.currenthamster.verbose())
        H2D2.currenthamster.logln(" to "+c.wayhome);
      }

      if (mydist<INFINITEDISTANCE)
      for(int dir = MINDIR; dir < NUMDIR; dir++) {
        if (c.isWall(dir)) continue;
        if (c.isCliffDown(dir)) continue;
        if (!c.isKnown(dir)) continue;
        Cell n = c.getLink(dir);
        if (n==null) continue;
        int ndist = n.wayhome.distance;
        //int ndir = n.wayhome.direction;
        if (ndist > mydist+1) {

          if (DEBUGWAYHOME && H2D2.currenthamster.verbose())
          H2D2.currenthamster.log("Add to queue "+n+": updated from "+n.wayhome);

          n.wayhome.distance = mydist+1;
          n.wayhome.direction = ANTIDIR[dir];

          if (DEBUGWAYHOME && H2D2.currenthamster.verbose())
          H2D2.currenthamster.logln(" to "+n.wayhome);

          queue.addLast(n);
        }
      }

    } // while

    if (DEBUGWAYHOME && H2D2.currenthamster.verbose())
    H2D2.currenthamster.logln("map.updateWayHome("+start+") leaving.");

  } // updateWayHome


  boolean updateWayMin0(Cell c) {
    if (DEBUGWAYMIN && H2D2.currenthamster.verbose())
    H2D2.currenthamster.logln("Check "+c+", min: "+c.waymin);

    // eigene Heimatentfernung
    int mydist = c.waymin.distance;
    int mydir = c.waymin.direction;
    int newdist = INFINITEDISTANCE;
    int newdir = -1;
    if (mydist == 0) return false; // Heimatfeld, wird nicht veraendert

    if (mydir>=MINDIR && !c.isWall(mydir) && !c.isCliffUp(mydir)
    && mydist==1+c.getLink(mydir).waymin.distance)
      return false;  // keine Aenderungen

    // wie wissen nun also, dass der alte Minimalweg nicht begehbar ist
    // oder die Distanz sich geaendert hat

    // ermittelt die kleinste Minimalentfernung eines Nachbarfeldes,
    // das nicht dieses Feld als Minimalweg benutzt
    for(int dir = MINDIR; dir <= MAXDIR; dir++) {
      if (c.isWall(dir)) continue;
      if (c.isCliffUp(dir)) continue;
      //if (!c.isKnown(dir)) continue;
      Cell n = c.getLink(dir);
      if (n==null) continue;
      int ndist = n.waymin.distance;
      int ndir = n.waymin.direction;

      if (DEBUGWAYMIN && H2D2.currenthamster.verbose())
      H2D2.currenthamster.logln("Neighbor "+n+", min:"+n.waymin);

      if (ndist < newdist && ndir != ANTIDIR[dir]) {
        newdist = ndist; newdir = dir;
      }
    }

    if (DEBUGWAYMIN && H2D2.currenthamster.verbose())
    H2D2.currenthamster.log("Update "+c+" from "+c.waymin);

    if (newdist == INFINITEDISTANCE) {
      // kein Ziel-Nachbar gefunden
      mydist = c.waymin.distance = INFINITEDISTANCE;
      mydir = c.waymin.direction = -1;
    }else{
      mydist = c.waymin.distance = newdist+1;
      mydir = c.waymin.direction = newdir;
    }

    if (DEBUGWAYMIN && H2D2.currenthamster.verbose())
    H2D2.currenthamster.logln(" to "+c.waymin);

    return true;
  }

  void updateWayMin(Cell start) {
    if (DEBUGWAYMIN && H2D2.currenthamster.verbose())
    H2D2.currenthamster.logln("map.updateWayMin("+start+") entered.");

    PriorityQueue queue = new PriorityQueue(CompareWayMin.COMPARATOR);

    queue.addFirst(start);
    while (!queue.isEmpty()) {
      Cell c = (Cell)queue.removeMin();

      if (DEBUGWAYMIN && H2D2.currenthamster.verbose())
      H2D2.currenthamster.logln("Work at "+c+", min: "+c.waymin);

      boolean changed = updateWayMin0(c);
      int mydist = c.waymin.distance;

      if (DEBUGWAYMIN && H2D2.currenthamster.verbose())
      H2D2.currenthamster.logln("Check neighbors at "+c);

      for(int dir = MINDIR; dir <= MAXDIR; dir++) {
        Cell n = c.getLink(dir);
        if (n==null) continue;
        int ndist = n.waymin.distance;
        int ndir = n.waymin.direction;

        // Minimalweg von n fuehrt ueber dieses Feld und muss geaendert werden
        if ((ndir == ANTIDIR[dir] && (changed || c.isWall(dir) || c.isCliffDown(dir)
        || ndist != mydist+1))) {

          //updateWayMin0(n);
          n.waymin.distance = INFINITEDISTANCE;
          n.waymin.direction = -1;

          if (ndist != n.waymin.distance) {
            if (DEBUGWAYMIN && H2D2.currenthamster.verbose())
            H2D2.currenthamster.logln("Add to queue "+n);
            queue.addLast(n);
          }
        }
      }

    } // while

    if (DEBUGWAYMIN && H2D2.currenthamster.verbose())
    H2D2.currenthamster.logln("map.updateWayMin("+start+") leaving.");

  } // updateWayMin

} // Map
